from . __class_init__ import *


class QoSElement(cdt.Structure):
    """ Params of element """
    precedence: cdt.Unsigned
    delay: cdt.Unsigned
    reliability: cdt.Unsigned
    peak_throughput: cdt.Unsigned
    mean_throughput: cdt.Unsigned


class QualityOfService(cdt.Structure):
    """ Specifies the quality of service parameters. It is a structure of 2 elements:
            1: defines the default or minimum characteristics of the network concerned. These parameters have to be set to best effort value;
            2: defines the requested parameters. """
    default: QoSElement
    requested: QoSElement


class GPRSModemSetup(ic.COSEMInterfaceClasses):
    """ This IC allow setting up GPRS modems, by handling all data necessary data for modem management. """
    NAME = cn.GPRS_MODEM_SETUP
    CLASS_ID = ClassID.GPRS_MODEM_SETUP
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement(an.APN, cdt.OctetString),
                  ic.ICAElement(an.PIN_CODE, cdt.LongUnsigned),
                  ic.ICAElement(an.QUALITY_OF_SERVICE, QualityOfService))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def APN(self) -> cdt.OctetString:
        return self.get_attr(2)

    @property
    def PIN_code(self) -> cdt.LongUnsigned:
        return self.get_attr(3)

    @property
    def quality_of_service(self) -> QualityOfService:
        return self.get_attr(4)



MANUAL = "Manual. By button"
LOCAL_1 = "Local. Active Power"
LOCAL_2 = "Local. Current"
LOCAL_3 = "Local. Voltage"
LOCAL_4 = "Local. Magnetic field"
LOCAL_5 = "Local. Current unbalance"
LOCAL_6 = "Local. Temperature"
LOCAL_7 = "Local. Register monitor"



from ..cosem_interface_classes import cosem_interface_class as ic, attr_indexes as ai
from ..cosem_interface_classes.overview import ClassID, Version
from ..types import common_data_types as cdt, cosem_service_types as cst, useful_types as ut
from ..settings import get_current_language, Language

match get_current_language():
    case Language.ENGLISH: from ..Values.EN import class_names as cn, attr_names as an, meth as mn
    case Language.RUSSIAN: from ..Values.RU import class_names as cn, attr_names as an, meth as mn



from typing import Any
from .collection import ic, cdt
from ..cosem_interface_classes import collection
from ..config_parser import config, get_values


def get_obj_report(
        obj: ic.COSEMInterfaceClasses,
        attr_index_par: tuple[int | Any, ...]) -> str:
    struct_report: dict | None = get_values("DLMS", "report", "struct")
    ret = str()
    ret += F"[{collection.get_name(obj.logical_name)}]\n"
    for i in attr_index_par:
        if isinstance(i, int):
            pass
        else:
            i, par = i[0], i[1:]
        value = obj.get_attr(i)
        if isinstance(value, cdt.SimpleDataType):
            match obj, i:
                case collection.impl.data.DLMSDeviceIDObject(), 2: value = value.to_str()
                case _: pass
            if isinstance(value, cdt.OctetString):
                pass
            if hasattr(value, "report"):
                value = value.report
            ret += F"  {obj.get_attr_element(i).NAME}: {value}\n"
        elif isinstance(value, cdt.ComplexDataType):
            ret += F"  [{obj.get_attr_element(i).NAME}]\n"
            stack: list = [("", iter(value))]
            while stack:
                name, value_it = stack[-1]
                indent = F"{' '*(len(stack) + 1)}"
                value = next(value_it, None)
                if value:
                    if not isinstance(name, str):
                        name = next(name).NAME
                    if isinstance(value, cdt.Array):
                        ret += F"{indent}[{name}]\n"
                        stack.append(("*", iter(value)))
                    elif isinstance(value, cdt.Structure):
                        if struct_report and (pattern := struct_report.get(value.__class__.__name__)):
                            val = list(pattern)
                            val.reverse()
                            result = str()
                            while val:
                                i = val.pop()
                                match i:
                                    case "%":
                                        par = val.pop()
                                        index = int(val.pop() + val.pop())
                                        match par:
                                            case "n":
                                                result += value.ELEMENTS[index].NAME
                                            case "v":
                                                result += str(value[index])
                                            case err:
                                                raise ValueError(F"unknown macros &{err}{index}")
                                    case symbol:
                                        result += symbol
                            ret += F"{indent}{result}\n"
                        else:
                            if name == "":
                                ret += "\n"
                            else:
                                ret += F"{indent}[{name}]\n"
                            stack.append((iter(value.ELEMENTS), iter(value)))
                    else:
                        ret += F"{indent}{name}: {value}\n"
                else:
                    stack.pop()
    return ret



from .__class_init__ import *
from ..types.implementations import structs, integers


class Actions(cdt.Array):
    """ Array of key_data """
    TYPE = structs.ActionItem


class ActorPermissions(cdt.BitString):
    """TODO: """


class PermissionsTable(cdt.Array):
    TYPE = ActorPermissions
    __iter__: ActorPermissions
    __get_item__: ActorPermissions


class ActorActionWeight(cdt.LongUnsigned):
    """TODO: make any thing"""


class ActorWeightingList(cdt.Array):
    TYPE = ActorActionWeight


class WeightingsTable(cdt.Array):
    TYPE = ActorWeightingList


class MostRecentRequest(cdt.BitString):
    """TODO: """


class MostRecentRequestTable(cdt.Array):
    TYPE = MostRecentRequest


class RequestAction(cdt.Structure):
    """Defines the actions that are requested by an actorDefines the actions that are requested by an actor"""
    request_actor: cdt.Unsigned
    request_action_list: cdt.BitString


class Arbitrator(ic.COSEMInterfaceClasses):
    """DLMS UA 1000-1 Ed. 14 4.5.12 Arbitrator"""
    NAME = cn.ARBITRATOR
    CLASS_ID = ClassID.ARBITRATOR
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement(an.ACTIONS, Actions),
                  ic.ICAElement(an.PERMISSION_TABLE, PermissionsTable),
                  ic.ICAElement(an.WEIGHTINGS_TABLE, WeightingsTable),
                  ic.ICAElement(an.MOST_RECENT_REQUESTS_TABLE, MostRecentRequestTable, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.LAST_OUTCOME, cdt.Unsigned, min=0, default=0, classifier=ic.Classifier.DYNAMIC))  # TODO: max = n what it?

    M_ELEMENTS = (ic.ICMElement(mn.REQUEST_ACTION, RequestAction),
                  ic.ICMElement(mn.RESET, integers.Only0))
    actors: tuple[str] = tuple()
    """name actors container"""

    def characteristics_init(self):
        self.set_attr(2, None)
        self.actions.register_cb_post_set(self.__check_permission_table)
        self.set_attr(3, None)
        self.set_attr(4, None)
        self.set_attr(5, None)

    @property
    def actions(self) -> Actions:
        return self.get_attr(2)

    @property
    def permissions_table(self) -> PermissionsTable:
        return self.get_attr(3)

    @property
    def weightings_table(self) -> WeightingsTable:
        return self.get_attr(4)

    @property
    def most_recent_request_table(self) -> MostRecentRequestTable:
        return self.get_attr(5)

    @property
    def last_outcome(self) -> cdt.Unsigned:
        return self.get_attr(6)

    @property
    def request_action(self) -> RequestAction:
        return self .get_meth(1)

    @property
    def reset(self) -> integers.Only0:
        return self .get_meth(2)

    def __check_permission_table(self):
        """set length actor_permission be same as action array size if it not valid"""
        if len(self.permissions_table) > 0:
            for actor in self.permissions_table:
                if len(actor) != len(self.actions):
                    actor.set(ActorPermissions('0'*len(self.actions)))
                else:
                    """lenght is correct, change is not required """
        else:
            """not was loaded"""

    def get_action_names(self) -> list[str]:
        """return names of methods by ordering actions"""
        return [self.collection.get_script_names(ln, selector) for ln, selector in self.actions]



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview
from src.DLMS_SPODES.version import AppVersion


class TestType(unittest.TestCase):

    def test_ScalerUnitType(self):
        su = cdt.ScalUnitType((0, 27))
        print(su)
        value = cdt.DoubleLong(1.2, scaler_unit=su)
        a = str(value)
        print(value, value.report)



from enum import IntEnum
from .enums import Transmit, Application
from .version import AppVersion
from . import pdu_enums as pdu


class TomlKeyError(Exception):
    """for handle exceptions of toml"""


class DLMSException(Exception):
    """ Common InterTechElectric exceptions class """
    error: Transmit | Application = Transmit.UNKNOWN


class AssociationResultError(DLMSException):
    """ Result of the proposed AA establishment and eventually the reason of the rejection of the association establishment request, as it is specified in ISO/IEC 8650-1.
        When no diagnostics are included, a null value is assigned to the result-source-diagnostics field. IEC62056-53 2004 7.3.3 The AARQ and AARE APDUs """
    error = Transmit.NO_ACCESS
    __match_args_ = ('result_source_diagnostics', )

    # TODO: make arg right type
    def __init__(self, result_source_diagnostics: IntEnum):
        Exception.__init__(self, F'Connection is {result_source_diagnostics.name}')
        self.result_source_diagnostics = result_source_diagnostics


class Timeout(DLMSException):
    """ timeout during connection or exchange """
    error = Transmit.TIMEOUT


class NoPort(DLMSException):
    """ not found port """
    error = Transmit.NO_PORT


class Abort(DLMSException):
    """ manual interrupt """
    error = Transmit.ABORT


class ITEConnection(DLMSException):
    """"""
    error = Transmit.UNKNOWN


class NoTransport(ITEConnection):
    """"""
    error = Transmit.NO_TRANSPORT


class ITEApplication(DLMSException):
    """"""


class NoObject(ITEApplication):
    """ object missing in collection """
    error = Application.MISSING_OBJ


class IDError(ITEApplication):
    """"""
    error = Application.ID_ERROR


class EmptyObj(ITEApplication):
    """ emtpy field in DLMS object """
    error = Application.EMPTY_OBJ


class NoConfig(ITEApplication):
    """ configuration for device not founded """
    error = Application.NO_CONFIG


class TypeErr(ITEApplication):
    """ unknown device type """
    error = Application.TYPE_ERROR


class VersionError(ITEApplication):
    """ Version error """
    error = Application.VERSION_ERROR

    def __init__(self, error_version: AppVersion, additional: str = 'device'):
        Exception.__init__(self, F'Unsupported {additional} version: {error_version}')
        self.version = error_version


class ResultError(ITEApplication):
    """ DLMS COSEMpdu_GB83.asn error """
    error = Application.RESULT_ERROR

    def __init__(self, error: pdu.DataAccessResult | pdu.ActionResult, additional: str = ""):
        Exception.__init__(self, F'{error.__class__.__name__}: {error.name} {additional}')
        self.result = error


class UnknownError(DLMSException):
    """ for unknown errors """
    error = Transmit.UNKNOWN


class NeedUpdate(ITEApplication):
    """error until there is no action"""
    error = Application.VERSION_ERROR



"""DLMS UA 1000-1 Ed 14 4.2 Overview of the COSEM interface classes"""
from enum import IntEnum
from dataclasses import dataclass
from ..types import ut, cdt


@dataclass(frozen=True)
class ClassID:
    DATA = ut.CosemClassId(1)
    REGISTER = ut.CosemClassId(3)
    EXT_REGISTER = ut.CosemClassId(4)
    DEMAND_REGISTER = ut.CosemClassId(5)
    REGISTER_ACTIVATION = ut.CosemClassId(6)
    PROFILE_GENERIC = ut.CosemClassId(7)
    CLOCK = ut.CosemClassId(8)
    SCRIPT_TABLE = ut.CosemClassId(9)
    SCHEDULE = ut.CosemClassId(10)
    SPECIAL_DAYS_TABLE = ut.CosemClassId(11)
    ASSOCIATION_SN = ut.CosemClassId(12)
    ASSOCIATION_LN = ut.CosemClassId(15)
    SAP_ASSIGNMENT = ut.CosemClassId(17)
    IMAGE_TRANSFER = ut.CosemClassId(18)
    IEC_LOCAL_PORT_SETUP = ut.CosemClassId(19)
    ACTIVITY_CALENDAR = ut.CosemClassId(20)
    REGISTER_MONITOR = ut.CosemClassId(21)
    SINGLE_ACTION_SCHEDULE = ut.CosemClassId(22)
    IEC_HDLC_SETUP = ut.CosemClassId(23)
    IEC_TWISTED_PAIR__1__SETUP = ut.CosemClassId(24)
    M_BUS_SLAVE_PORT_SETUP = ut.CosemClassId(25)
    UTILITY_TABLES = ut.CosemClassId(26)
    MODEM_CONFIGURATION = ut.CosemClassId(27)
    AUTO_ANSWER = ut.CosemClassId(28)
    AUTO_CONNECT = ut.CosemClassId(29)
    COSEM_DATA_PROTECTION = ut.CosemClassId(30)
    PUSH_SETUP = ut.CosemClassId(40)
    TCP_UDP_SETUP = ut.CosemClassId(41)
    IPV4_SETUP = ut.CosemClassId(42)
    PRIME_NB_OFDM_PLC_MAC_ADDRESS_SETUP = ut.CosemClassId(43)
    PPP_SETUP = ut.CosemClassId(44)
    GPRS_MODEM_SETUP = ut.CosemClassId(45)
    SMTP_SETUP = ut.CosemClassId(46)
    GSM_DIAGNOSTIC = ut.CosemClassId(47)
    IPV6_SETUP = ut.CosemClassId(48)
    S_FSK_PHY_MAC_SET_UP = ut.CosemClassId(50)
    S_FSK_ACTIVE_INITIATOR = ut.CosemClassId(51)
    S_FSK_MAC_SYNCHRONIZATION_TIMEOUTS = ut.CosemClassId(52)
    S_FSK_MAC_COUNTERS = ut.CosemClassId(53)
    IEC_61334_4_32_LLC_SETUP = ut.CosemClassId(55)
    S_FSK_REPORTING_SYSTEM_LIST = ut.CosemClassId(56)
    ISO_IEC_8802_2_LLC_TYPE_1_SETUP = ut.CosemClassId(57)
    ISO_IEC_8802_2_LLC_TYPE_2_SETUP = ut.CosemClassId(58)
    ISO_IEC_8802_2_LLC_TYPE_3_SETUP = ut.CosemClassId(59)
    REGISTER_TABLE = ut.CosemClassId(61)
    COMPACT_DATA = ut.CosemClassId(62)
    STATUS_MAPPING = ut.CosemClassId(63)
    SECURITY_SETUP = ut.CosemClassId(64)
    PARAMETER_MONITOR = ut.CosemClassId(65)
    SENSOR_MANAGER_INTERFACE_CLASS = ut.CosemClassId(67)
    ARBITRATOR = ut.CosemClassId(68)
    DISCONNECT_CONTROL = ut.CosemClassId(70)
    LIMITER = ut.CosemClassId(71)
    M_BUS_CLIENT = ut.CosemClassId(72)
    WIRELESS_MODE_Q_CHANNEL = ut.CosemClassId(73)
    M_BUS_MASTER_PORT_SETUP = ut.CosemClassId(74)
    DLMS_COSEM_SERVER_M_BUS_PORT_SETUP = ut.CosemClassId(76)
    M_BUS_DIAGNOSTIC = ut.CosemClassId(77)
    _61334_4_32_LLC_SSCS_SETUP = ut.CosemClassId(80)
    PRIME_NB_OFDM_PLC_PHYSICAL_LAYER_COUNTERS = ut.CosemClassId(81)
    PRIME_NB_OFDM_PLC_MAC_SETUP = ut.CosemClassId(82)
    PRIME_NB_OFDM_PLC_MAC_FUNCTIONAL_PARAMETERS = ut.CosemClassId(83)
    PRIME_NB_OFDM_PLC_MAC_COUNTERS = ut.CosemClassId(84)
    PRIME_NB_OFDM_PLC_MAC_NETWORK_ADMINISTRATION_DATA = ut.CosemClassId(85)
    PRIME_NB_OFDM_PLC_APPLICATION_IDENTIFICATION = ut.CosemClassId(86)
    G3_PLC_MAC_LAYER_COUNTERS = ut.CosemClassId(90)
    G3_PLC_MAC_SETUP = ut.CosemClassId(91)
    G3_PLC_MAC_6LOWPAN_ADAPTATION_LAYER_SETUP = ut.CosemClassId(92)
    WI_SUN_SETUP = ut.CosemClassId(95)
    WI_SUN_DIAGNOSTIC = ut.CosemClassId(96)
    RPL_DIAGNOSTIC = ut.CosemClassId(97)
    MPL_DIAGNOSTIC = ut.CosemClassId(98)
    NTP_SETUP = ut.CosemClassId(100)
    ZIGBEE_SAS_STARTUP = ut.CosemClassId(101)
    ZIGBEE_SAS_JOIN = ut.CosemClassId(102)
    ZIGBEE_SAS_APS_FRAGMENTATION = ut.CosemClassId(103)
    ZIGBEE_NETWORK_CONTROL = ut.CosemClassId(104)
    ZIGBEE_TUNNEL_SETUP = ut.CosemClassId(105)
    ACCOUNT = ut.CosemClassId(111)
    CREDIT_INTERFACE_CLASS = ut.CosemClassId(112)
    CHARGE = ut.CosemClassId(113)
    TOKEN_GATEWAY = ut.CosemClassId(115)
    FUNCTION_CONTROL = ut.CosemClassId(122)
    ARRAY_MANAGER = ut.CosemClassId(123)
    COMMUNICATION_PORT_PROTECTION = ut.CosemClassId(124)
    SCHC_LPWAN_SETUP = ut.CosemClassId(126)
    SCHC_LPWAN_DIAGNOSTIC = ut.CosemClassId(127)
    LoRaWAN_SETUP = ut.CosemClassId(128)
    LoRaWAN_DIAGNOSTIC = ut.CosemClassId(129)
    ISO_IEC14908_IDENTIFICATION = ut.CosemClassId(130)
    ISO_IEC14908_PROTOCOL_SETUP = ut.CosemClassId(131)
    ISO_IEC14908_PROTOCOL_STATUS = ut.CosemClassId(132)
    ISO_IEC14908_PROTOCOL_DIAGNOSTIC = ut.CosemClassId(133)
    HS_PLC_ISO_IEC_12139_1_MAC_SETUP = ut.CosemClassId(140)
    HS_PLC_ISO_IEC_12139_1_CPAS_SETUP = ut.CosemClassId(141)
    HS_PLC_ISO_IEC_12139_1_IP_SSAS_SETUP = ut.CosemClassId(142)
    HS_PLC_ISO_IEC_12139_1_HDLC_SSAS_SETUP = ut.CosemClassId(143)
    LTE_MONITORING = ut.CosemClassId(151)
    CLIENT_SETUP = ut.CosemClassId(32767)  # TODO: remove in future

    @classmethod
    def get_all_id(cls) -> tuple[int]:
        """return all id container in build-in <int>"""
        return tuple(map(int, filter(lambda it: isinstance(it, ut.CosemClassId), cls.__dict__.values())))


@dataclass(frozen=True)
class Version:
    V0 = cdt.Unsigned(0)
    V1 = cdt.Unsigned(1)
    V2 = cdt.Unsigned(2)
    V3 = cdt.Unsigned(3)


class CountrySpecificIdentifiers(IntEnum):
    FINLAND = 0
    USA = 1
    CANADA = 2
    SERBIA = 3
    RUSSIA = 7
    CZECH_REPUBLIC = 10
    BULGARIA = 11
    CROATIA = 12
    IRELAND = 13
    ISRAEL = 14
    UKRAINE = 15
    YUGOSLAVIA = 16
    EGYPT = 20
    SOUTH_AFRICA = 27
    GREECE = 30
    NETHERLANDS = 31
    BELGIUM = 32
    FRANCE = 33
    SPAIN = 34
    PORTUGAL = 35
    HUNGARY = 36
    LITHUANIA = 37
    SLOVENIA = 38
    ITALY = 39
    ROMANIA = 40
    SWITZERLAND = 41
    SLOVAKIA = 42
    AUSTRIA = 43
    UNITED_KINGDOM = 44
    DENMARK = 45
    SWEDEN = 46
    NORWAY = 47
    POLAND = 48
    GERMANY = 49
    PERU = 51
    SOUGH_KOREA = 52
    CUBA = 53
    ARGENTINA = 54
    BRAZIL = 55
    CHILE = 56
    COLOMBIA = 57
    VENEZUELA = 58
    MALAYSIA = 60
    AUSTRALIA = 61
    INDONESIA = 62
    PHILIPPINES = 63
    NEW_ZEALAND = 64
    SINGAPORE = 65
    THAILAND = 66
    LATVIA = 71
    MOLDOVA = 73
    BELARUS = 75
    JAPAN = 81
    MEXICO = 82
    HONG_KONG = 85
    CHINA = 86
    BOSNIA_AND_HERZEGOVINA = 87
    TURKEY = 90
    INDIA = 91
    PAKISTAN = 92
    SAUDI_ARABIA = 96
    UNITED_ARAB_EMIRATES = 97
    IRAN = 98

    def __str__(self):
        return self.name



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview
from src.DLMS_SPODES.cosem_interface_classes import implementations as impl
from src.DLMS_SPODES.version import AppVersion


class TestType(unittest.TestCase):

    def test_Data(self):
        obj = collection.Data("0.0.0.1.0.255")
        obj.set_attr(2, cdt.Unsigned(8).encoding)
        a = obj.get_attr(2)
        a.set(3)
        print(obj.value)

    def test_eventsData(self):
        col = collection.Collection()
        col.add(class_id=ut.CosemClassId(7), version=cdt.Unsigned(1), logical_name=cst.LogicalName("0.0.99.98.4.255"))
        print(col)

    def test_ExternalEventData(self):
        col = collection.Collection()
        col.set_manufacturer(b'KPZ')
        col.set_country("3.0")
        col.server_ver = AppVersion(1, 3, 0)
        col.server_type = cdt.OctetString("4d324d5f33")
        col.set_spec()
        col.add(class_id=ut.CosemClassId(1), version=cdt.Unsigned(0), logical_name=cst.LogicalName("0.0.96.11.4.255"))
        obj = col.get_object("0.0.96.11.4.255")
        obj.set_attr(2, b'\x06\x00\x00\x00\x02')
        # obj.set_attr(2, 2)
        print(obj.value, obj.value.report)
        self.assertEqual(obj.value.report, "Магнитное поле - окончание(2)", "report match")
        print(col)

    def test_DeviceIdObjects(self):
        obj = impl.data.DLMSDeviceIDObject("0.0.96.1.4.255")
        datatime = cdt.DateTime("01.01.20 11:00")
        obj.set_attr(2, datatime.encoding)

    def test_SealStatus(self):
        obj = impl.data.SealStatus("0.0.96.51.5.255")
        obj.set_attr(2, 0)
        for i in range(255):
            obj.value.set(i)
            print(i, obj.value.get_report())
